/*
 * Copyright (c) 2018 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-BSD-5-Clause-Nordic
 */

#include <zephyr.h>
#include <stdio.h>
#include <uart.h>
#include <string.h>

#include <net/mqtt.h>
#include <net/socket.h>
#include <lte_lc.h>

#include "cJSON.h"
#include "gps_controller.h"


#if defined(CONFIG_LWM2M_CARRIER)
#include <lwm2m_carrier.h>
#endif

#define  GPS_PROCESS_THREAD_STACK        1024
#define  GPS_THREAD_PROCESS_PRIORITY  11

#define  JSON_PACKAGE_ID                 "packID"
#define  JSON_GPS_PVT_OBJECT                 "gps_pvt"
#define  JSON_LATITUDE_ITEM                      "latitude"
#define  JSON_LONGITUDE_ITEM                  "longitude"
#define  JSON_ALTITUDE_ITEM                       "altitude"
#define  JSON_ACCURACY_ITEM                     "accuracy"
#define  JSON_SPEED_ITEM                     		 "speed"
#define  JSON_HEADING_ITEM                     	 "heading"
#define  JSON_DATE_ITEM                     	       "date"

#define  ASCII_CHAR_SLASH                              "/"



/* Buffers for MQTT client. */
static u8_t rx_buffer[CONFIG_MQTT_MESSAGE_BUFFER_SIZE];
static u8_t tx_buffer[CONFIG_MQTT_MESSAGE_BUFFER_SIZE];
static u8_t payload_buf[CONFIG_MQTT_PAYLOAD_BUFFER_SIZE];
static char publish_payload_buf[CONFIG_MQTT_PAYLOAD_BUFFER_SIZE];

/* The mqtt client struct */
static struct mqtt_client client;

/* MQTT Broker details. */
static struct sockaddr_storage broker;

/* Connected flag */
static bool mqtt_connected;
K_SEM_DEFINE(sem_net_ready, 0, 1);

/* File descriptor */
static struct pollfd fds;

/* Structures for work */
//static struct k_delayed_work send_gps_data_work;
//static struct k_delayed_work disconnect_mqtt_work;

static struct gps_data gps_data;
K_SEM_DEFINE(sem_gps_fix, 0, 1);

#if defined(CONFIG_BSD_LIBRARY)

/**@brief Recoverable BSD library error. */
void bsd_recoverable_error_handler(uint32_t err)
{
	printk("bsdlib recoverable error: %u\n", (unsigned int)err);
}

#endif /* defined(CONFIG_BSD_LIBRARY) */

#if defined(CONFIG_LWM2M_CARRIER)
K_SEM_DEFINE(carrier_registered, 0, 1);

void lwm2m_carrier_event_handler(const lwm2m_carrier_event_t *event)
{
	switch (event->type) {
	case LWM2M_CARRIER_EVENT_BSDLIB_INIT:
		printk("LWM2M_CARRIER_EVENT_BSDLIB_INIT\n");
		break;
	case LWM2M_CARRIER_EVENT_CONNECT:
		printk("LWM2M_CARRIER_EVENT_CONNECT\n");
		break;
	case LWM2M_CARRIER_EVENT_DISCONNECT:
		printk("LWM2M_CARRIER_EVENT_DISCONNECT\n");
		break;
	case LWM2M_CARRIER_EVENT_READY:
		printk("LWM2M_CARRIER_EVENT_READY\n");
		k_sem_give(&carrier_registered);
		break;
	case LWM2M_CARRIER_EVENT_FOTA_START:
		printk("LWM2M_CARRIER_EVENT_FOTA_START\n");
		break;
	case LWM2M_CARRIER_EVENT_REBOOT:
		printk("LWM2M_CARRIER_EVENT_REBOOT\n");
		break;
	}
}
#endif /* defined(CONFIG_LWM2M_CARRIER) */

/**@brief Function to print strings without null-termination
 */
static void data_print(u8_t *prefix, u8_t *data, size_t len)
{
	char buf[len + 1];

	memcpy(buf, data, len);
	buf[len] = 0;
	printk("%s%s\n", prefix, buf);
}

/**@brief Function to publish data on the configured topic
 */
static int data_publish(struct mqtt_client *c, enum mqtt_qos qos,
	u8_t *data, size_t len)
{
	struct mqtt_publish_param param;

	param.message.topic.qos = qos;
	param.message.topic.topic.utf8 = CONFIG_MQTT_PUB_TOPIC;
	param.message.topic.topic.size = strlen(CONFIG_MQTT_PUB_TOPIC);
	param.message.payload.data = data;
	param.message.payload.len = len;
	param.message_id = sys_rand32_get();
	param.dup_flag = 0;
	param.retain_flag = 0;

	data_print("Publishing: ", data, len);
	printk("to topic: %s len: %u\n",
		CONFIG_MQTT_PUB_TOPIC,
		(unsigned int)strlen(CONFIG_MQTT_PUB_TOPIC));

	return mqtt_publish(c, &param);
}

static void send_gps_data_work_fn(void)
{
	static uint32_t package_id = 0;
	
	printk("send gps data to server\n");

	cJSON *m_object = cJSON_CreateObject();
	if (m_object == NULL)
	{
		cJSON_Delete(m_object); 	
		return;
	}

	package_id++;

	cJSON *gps_pvt_object = cJSON_AddObjectToObject(m_object, JSON_GPS_PVT_OBJECT); 

	cJSON_AddNumberToObject(m_object, JSON_PACKAGE_ID, package_id);

	cJSON_AddNumberToObject(gps_pvt_object, JSON_LATITUDE_ITEM, gps_data.pvt.latitude);
	// cJSON_AddNumberToObject(gps_pvt_object, JSON_LONGITUDE_ITEM, gps_data.pvt.longitude);
	// cJSON_AddNumberToObject(gps_pvt_object, JSON_ALTITUDE_ITEM, gps_data.pvt.altitude);
	// cJSON_AddNumberToObject(gps_pvt_object, JSON_ACCURACY_ITEM, gps_data.pvt.accuracy);
	// cJSON_AddNumberToObject(gps_pvt_object, JSON_SPEED_ITEM, gps_data.pvt.speed);
	// cJSON_AddNumberToObject(gps_pvt_object, JSON_HEADING_ITEM, gps_data.pvt.heading);

	// char date_str[64];
	// memset(date_str, 0, sizeof(date_str));
	// sprintf(date_str, "%d/%d/%d %d:%d:%d", gps_data.pvt.datetime.year, 
	// 	gps_data.pvt.datetime.month, 
	// 	gps_data.pvt.datetime.day,
	// 	gps_data.pvt.datetime.hour,
	// 	gps_data.pvt.datetime.minute,
	// 	gps_data.pvt.datetime.seconds);

	// cJSON_AddStringToObject(gps_pvt_object, JSON_DATE_ITEM, date_str);
	
	char *ret = cJSON_Print(m_object);
	if ( ret == NULL) 
	{
		goto end;     
	}

	if(mqtt_connected)
	{
		memset(publish_payload_buf, 0, sizeof(publish_payload_buf));
		memcpy(publish_payload_buf, ret, strlen(ret));
		data_publish(&client, MQTT_QOS_1_AT_LEAST_ONCE, publish_payload_buf, strlen(publish_payload_buf));		
	}
	cJSON_free(ret);

	end:
	cJSON_Delete(m_object); 	
	//cJSON_Delete(gps_pvt_object); 
}

/**@brief Function to subscribe to the configured topic
 */
static int subscribe(void)
{
	struct mqtt_topic subscribe_topic = {
		.topic = {
			.utf8 = CONFIG_MQTT_SUB_TOPIC,
			.size = strlen(CONFIG_MQTT_SUB_TOPIC)
		},
		.qos = MQTT_QOS_1_AT_LEAST_ONCE
	};

	const struct mqtt_subscription_list subscription_list = {
		.list = &subscribe_topic,
		.list_count = 1,
		.message_id = 1234
	};

	printk("Subscribing to: %s len %u\n", CONFIG_MQTT_SUB_TOPIC,
		(unsigned int)strlen(CONFIG_MQTT_SUB_TOPIC));

	return mqtt_subscribe(&client, &subscription_list);
}

/**@brief Function to read the published payload.
 */
static int publish_get_payload(struct mqtt_client *c, size_t length)
{
	u8_t *buf = payload_buf;
	u8_t *end = buf + length;

	if (length > sizeof(payload_buf)) {
		return -EMSGSIZE;
	}

	while (buf < end) {
		int ret = mqtt_read_publish_payload(c, buf, end - buf);

		if (ret < 0) {
			int err;

			if (ret != -EAGAIN) {
				return ret;
			}

			printk("mqtt_read_publish_payload: EAGAIN\n");

			err = poll(&fds, 1, K_SECONDS(CONFIG_MQTT_KEEPALIVE));
			if (err > 0 && (fds.revents & POLLIN) == POLLIN) {
				continue;
			} else {
				return -EIO;
			}
		}

		if (ret == 0) {
			return -EIO;
		}

		buf += ret;
	}

	return 0;
}

/**@brief MQTT client event handler
 */
void mqtt_evt_handler(struct mqtt_client *const c,
		      const struct mqtt_evt *evt)
{
	int err;

	switch (evt->type) {
	case MQTT_EVT_CONNACK:
		if (evt->result != 0) {
			printk("MQTT connect failed %d\n", evt->result);
			break;
		}

		mqtt_connected = true;
		printk("[%s:%d] MQTT client connected!\n", __func__, __LINE__);
		subscribe();
		 k_sem_give(&sem_net_ready);
		break;

	case MQTT_EVT_DISCONNECT:
		printk("[%s:%d] MQTT client disconnected %d\n", __func__,
		       __LINE__, evt->result);
		mqtt_connected = false;
		break;

	case MQTT_EVT_PUBLISH: {
		const struct mqtt_publish_param *p = &evt->param.publish;

		printk("[%s:%d] MQTT PUBLISH result=%d len=%d\n", __func__,
		       __LINE__, evt->result, p->message.payload.len);
		err = publish_get_payload(c, p->message.payload.len);
		if (err >= 0) {
			data_print("Received: ", payload_buf,
				p->message.payload.len);
			/* Echo back received data */
			data_publish(&client, MQTT_QOS_1_AT_LEAST_ONCE,
				payload_buf, p->message.payload.len);
		} else {
			printk("mqtt_read_publish_payload: Failed! %d\n", err);
			printk("Disconnecting MQTT client...\n");

			err = mqtt_disconnect(c);
			if (err) {
				printk("Could not disconnect: %d\n", err);
			}
		}
	} break;

	case MQTT_EVT_PUBACK:
		if (evt->result != 0) {
			printk("MQTT PUBACK error %d\n", evt->result);
			break;
		}

		printk("[%s:%d] PUBACK packet id: %u\n", __func__, __LINE__,
				evt->param.puback.message_id);
		break;

	case MQTT_EVT_SUBACK:
		if (evt->result != 0) {
			printk("MQTT SUBACK error %d\n", evt->result);
			break;
		}

		printk("[%s:%d] SUBACK packet id: %u\n", __func__, __LINE__,
				evt->param.suback.message_id);
		break;

	default:
		printk("[%s:%d] default: %d\n", __func__, __LINE__,
				evt->type);
		break;
	}
}

/**@brief Resolves the configured hostname and
 * initializes the MQTT broker structure
 */
static void broker_init(void)
{
	int err;
	struct addrinfo *result;
	struct addrinfo *addr;
	struct addrinfo hints = {
		.ai_family = AF_INET,
		.ai_socktype = SOCK_STREAM
	};

	err = getaddrinfo(CONFIG_MQTT_BROKER_HOSTNAME, NULL, &hints, &result);
	if (err) {
		printk("ERROR: getaddrinfo failed %d\n", err);

		return;
	}

	addr = result;
	err = -ENOENT;

	/* Look for address of the broker. */
	while (addr != NULL) {
		/* IPv4 Address. */
		if (addr->ai_addrlen == sizeof(struct sockaddr_in)) {
			struct sockaddr_in *broker4 =
				((struct sockaddr_in *)&broker);
			char ipv4_addr[NET_IPV4_ADDR_LEN];

			broker4->sin_addr.s_addr =
				((struct sockaddr_in *)addr->ai_addr)
				->sin_addr.s_addr;
			broker4->sin_family = AF_INET;
			broker4->sin_port = htons(CONFIG_MQTT_BROKER_PORT);

			inet_ntop(AF_INET, &broker4->sin_addr.s_addr,
				  ipv4_addr, sizeof(ipv4_addr));
			printk("IPv4 Address found %s\n", ipv4_addr);

			break;
		} else {
			printk("ai_addrlen = %u should be %u or %u\n",
				(unsigned int)addr->ai_addrlen,
				(unsigned int)sizeof(struct sockaddr_in),
				(unsigned int)sizeof(struct sockaddr_in6));
		};
	}

	/* Free the address. */
	freeaddrinfo(result);
}


static char mqtt_client_id[64];

/**@brief Initialize the MQTT client structure
 */
static void client_init(struct mqtt_client *client)
{
	mqtt_client_init(client);

	broker_init();

	/* MQTT client configuration */
	client->broker = &broker;
	client->evt_cb = mqtt_evt_handler;

	memset(mqtt_client_id, 0, sizeof(mqtt_client_id));
	memcpy(mqtt_client_id, CONFIG_MQTT_CLIENT_ID, strlen(CONFIG_MQTT_CLIENT_ID));
	sprintf(mqtt_client_id +strlen(mqtt_client_id), "%08x", sys_rand32_get());
	client->client_id.utf8 = (u8_t *)mqtt_client_id;
	client->client_id.size = strlen(mqtt_client_id);
	printk("client id =%s\n", mqtt_client_id);
	client->password = NULL;
	client->user_name = NULL;
	client->protocol_version = MQTT_VERSION_3_1_1;

	/* MQTT buffers configuration */
	client->rx_buf = rx_buffer;
	client->rx_buf_size = sizeof(rx_buffer);
	client->tx_buf = tx_buffer;
	client->tx_buf_size = sizeof(tx_buffer);

	/* MQTT transport configuration */
	client->transport.type = MQTT_TRANSPORT_NON_SECURE;
}

/**@brief Initialize the file descriptor structure used by poll.
 */
static int fds_init(struct mqtt_client *c)
{
	if (c->transport.type == MQTT_TRANSPORT_NON_SECURE) {
		fds.fd = c->transport.tcp.sock;
	} else {
#if defined(CONFIG_MQTT_LIB_TLS)
		fds.fd = c->transport.tls.sock;
#else
		return -ENOTSUP;
#endif
	}

	fds.events = POLLIN;

	return 0;
}

/**@brief Configures modem to provide LTE link. Blocks until link is
 * successfully established.
 */
static void modem_configure(void)
{
#if defined(CONFIG_LTE_LINK_CONTROL)
	if (IS_ENABLED(CONFIG_LTE_AUTO_INIT_AND_CONNECT)) {
		/* Do nothing, modem is already turned on
		 * and connected.
		 */
	} else {
#if defined(CONFIG_LWM2M_CARRIER)
		/* Wait for the LWM2M_CARRIER to configure the modem and
		 * start the connection.
		 */
		printk("Waitng for carrier registration...\n");
		k_sem_take(&carrier_registered, K_FOREVER);
		printk("Registered!\n");
#else /* defined(CONFIG_LWM2M_CARRIER) */
		int err;

		printk("LTE Link Connecting ...\n");
		err = lte_lc_init_and_connect();
		__ASSERT(err == 0, "LTE link could not be established.");
		printk("LTE Link Connected!\n");
#endif /* defined(CONFIG_LWM2M_CARRIER) */
	}
#endif /* defined(CONFIG_LTE_LINK_CONTROL) */
}

/**@brief Callback for GPS trigger events */
static void gps_trigger_handler(struct device *dev, struct gps_trigger *trigger)
{
	ARG_UNUSED(trigger);


	gps_sample_fetch(dev);
	gps_channel_get(dev, GPS_CHAN_PVT, &gps_data);
	//LOG_HEXDUMP_INF(gps_data.nmea.buf, gps_data.nmea.len, "gps data=");
	char gps_string[32];
	memset(gps_string, 0, sizeof(gps_string));
	sprintf(gps_string, "%f", (float)gps_data.pvt.latitude);
	printk("gps latitud = %s\n", gps_string);

	memset(gps_string, 0, sizeof(gps_string));
	sprintf(gps_string, "%f", (float)gps_data.pvt.longitude);
	printk("gps longitude =%s\n", gps_string);

	memset(gps_string, 0, sizeof(gps_string));
	sprintf(gps_string, "%f", gps_data.pvt.altitude );
	printk("gps altitude =%s\n", gps_string);

	memset(gps_string, 0, sizeof(gps_string));
	sprintf(gps_string, "%f", gps_data.pvt.accuracy);
	printk("gps accuracy =%s\n", gps_string);

	memset(gps_string, 0, sizeof(gps_string));
	sprintf(gps_string, "%f\n", gps_data.pvt.speed);
	printk("gps speed =%s", gps_string);

	memset(gps_string, 0, sizeof(gps_string));
	sprintf(gps_string, "%f",  gps_data.pvt.heading);
	printk("gps heading =%s\n", gps_string);
	printk("gps flags =%d\n", gps_data.pvt.flags );
	printk("gps flags =%d/%d/%d %d:%d:%d\n", gps_data.pvt.datetime.year, 
		gps_data.pvt.datetime.month, 
		gps_data.pvt.datetime.day,
		gps_data.pvt.datetime.hour,
		gps_data.pvt.datetime.minute,
		gps_data.pvt.datetime.seconds
		);
	//gps_control_stop(K_NO_WAIT);
	k_sem_give(&sem_gps_fix);	
	//k_delayed_work_submit(&send_gps_data_work, K_NO_WAIT);
}

/**@brief Initializes and submits delayed work. */
// static void work_init(void)
// {
// 	k_delayed_work_init(&send_gps_data_work, send_gps_data_work_fn);
// 	k_delayed_work_init(&disconnect_mqtt_work, disconnect_mqtt_work_fn);
// }
void main(void)
{
	int err;

	printk("The MQTT simple sample started\n");
	modem_configure();
	gps_control_init(gps_trigger_handler);
	gps_control_enable();	
	
	//gps_control_start(K_SECONDS(1));

	while(1)
	{
		//k_sleep(K_FOREVER);		
		printk("mqtt domain %s\n", CONFIG_MQTT_BROKER_HOSTNAME);
		client_init(&client);		
		err = mqtt_connect(&client);
		if (err != 0) {
			printk("ERROR: mqtt_connect %d\n", err);
		}

		err = fds_init(&client);
		if (err != 0) {
			printk("ERROR: fds_init %d\n", err);
		}

		while (1)
		{
			
			err = poll(&fds, 1, K_SECONDS(CONFIG_MQTT_KEEPALIVE));
			//err = poll(&fds, 1, K_SECONDS(3));
			if (err < 0) {
				printk("ERROR: poll %d\n", errno);
				break;
			}

			err = mqtt_live(&client);
			if (err != 0) {
				printk("ERROR: mqtt_live %d\n", err);
				break;
			}

			if ((fds.revents & POLLIN) == POLLIN) {
				err = mqtt_input(&client);
				if (err != 0) {
					printk("ERROR: mqtt_input %d\n", err);
					break;
				}
			}

			if ((fds.revents & POLLERR) == POLLERR) {
				printk("POLLERR\n");
				break;
			}

			if ((fds.revents & POLLNVAL) == POLLNVAL) {
				printk("POLLNVAL\n");
				break;
			}
		}

		printk("Disconnecting MQTT client...\n");

		err = mqtt_disconnect(&client);
		if (err) {
			printk("Could not disconnect MQTT client. Error: %d\n", err);
		}
	}
}

void gps_thread_process_fn(int unused1, int unused2, int unused3)
{
	printk(">>entry gps thread\n");
	//k_sem_take(&sem_net_ready, K_FOREVER);
    while (1) 
	{
		k_sem_take(&sem_gps_fix, K_FOREVER);
		// mqtt_ping(&client);
		// k_sleep(K_SECONDS(20));
		send_gps_data_work_fn();
    }
}


K_THREAD_DEFINE(gps_thread_id, GPS_PROCESS_THREAD_STACK, gps_thread_process_fn, 
NULL, NULL, NULL, 
GPS_THREAD_PROCESS_PRIORITY, 0, K_NO_WAIT);
